@@ -1,5 +1,5 @@
 /*
- * $Id: live365.c,v 1.42 2004/01/27 11:38:19 jylefort Exp $
+ * $Id: live365.c,v 1.48 2004/03/28 17:16:04 jylefort Exp $
  *
  * Copyright (c) 2002, 2003, 2004 Jean-Yves Lefort
  * All rights reserved.
@@ -54,9 +54,11 @@ typedef struct
   char		*description;
   char		*broadcaster;
   char		*audio;
-  
   char		*url_postfix;
-  char		*station_url;
+  char		*homepage;
+
+  char		*url;		/* will be read and written in threads... */
+  GMutex	*url_mutex;	/* ...so we protect it */
 } Live365Stream;
 
 typedef struct
@@ -80,7 +82,8 @@ enum {
   FIELD_BROADCASTER,
   FIELD_AUDIO,
   FIELD_URL_POSTFIX,
-  FIELD_STATION_URL,
+  FIELD_HOMEPAGE,
+  FIELD_URL
 };
 
 /*** variables ***************************************************************/
@@ -114,13 +117,21 @@ static void		stream_field_set_cb	(Live365Stream	*stream,
 						 STHandlerField	*field,
 						 const GValue	*value,
 						 gpointer	data);
+static void		stream_stock_field_get_cb (Live365Stream	*stream,
+						   STHandlerStockField	stock_field,
+						   GValue		*value,
+						   gpointer		data);
 static void		stream_free_cb		(Live365Stream	*stream,
 						 gpointer	data);
 
-static char		*stream_get_url            (Live365Stream *stream,
-						    GError        **err);
-static char		*streaminfo_get_station_id (const char    *streaminfo);
+static gboolean		stream_resolve		(Live365Stream	*stream,
+						 GError		**err);
+static void		stream_resolve_line_cb	(const char	*line,
+						 gpointer	data);
 
+static gboolean		stream_resolve_cb	(Live365Stream	*stream,
+						 gpointer	data,
+						 GError		**err);
 static gboolean		stream_tune_in_cb	(Live365Stream	*stream,
 						 gpointer	data,
 						 GError		**err);
@@ -162,7 +173,12 @@ static gboolean		search_url_cb		(STCategory	*category);
 static Live365Stream *
 stream_new_cb (gpointer data)
 {
-  return g_new0(Live365Stream, 1);
+  Live365Stream *stream;
+
+  stream = g_new0(Live365Stream, 1);
+  stream->url_mutex = g_mutex_new();
+
+  return stream;
 }
 
 static void
@@ -197,8 +213,14 @@ stream_field_get_cb (Live365Stream *stream,
       g_value_set_string(value, stream->url_postfix);
       break;
 
-    case FIELD_STATION_URL:
-      g_value_set_string(value, stream->station_url);
+    case FIELD_HOMEPAGE:
+      g_value_set_string(value, stream->homepage);
+      break;
+
+    case FIELD_URL:
+      g_mutex_lock(stream->url_mutex);
+      g_value_set_string(value, stream->url);
+      g_mutex_unlock(stream->url_mutex);
       break;
 
     default:
@@ -238,8 +260,14 @@ stream_field_set_cb (Live365Stream *stream,
       stream->url_postfix = g_value_dup_string(value);
       break;
 
-    case FIELD_STATION_URL:
-      stream->station_url = g_value_dup_string(value);
+    case FIELD_HOMEPAGE:
+      stream->homepage = g_value_dup_string(value);
+      break;
+
+    case FIELD_URL:
+      g_mutex_lock(stream->url_mutex);
+      stream->url = g_value_dup_string(value);
+      g_mutex_unlock(stream->url_mutex);
       break;
 
     default:
@@ -247,6 +275,24 @@ stream_field_set_cb (Live365Stream *stream,
     }
 }
 
+static void
+stream_stock_field_get_cb (Live365Stream *stream,
+			   STHandlerStockField stock_field,
+			   GValue *value,
+			   gpointer data)
+{
+  switch (stock_field)
+    {
+    case ST_HANDLER_STOCK_FIELD_NAME:
+      g_value_set_string(value, stream->title);
+      break;
+
+    case ST_HANDLER_STOCK_FIELD_GENRE:
+      g_value_set_string(value, stream->genre);
+      break;
+    }
+}
+
 static void
 stream_free_cb (Live365Stream *stream, gpointer data)
 {
@@ -256,66 +302,79 @@ stream_free_cb (Live365Stream *stream, gpointer data)
   g_free(stream->broadcaster);
   g_free(stream->audio);
   g_free(stream->url_postfix);
-  g_free(stream->station_url);
+  g_free(stream->homepage);
+
+  g_free(stream->url);
+  g_mutex_free(stream->url_mutex);
 
   st_stream_free((STStream *) stream);
 }
 
-static char *
-stream_get_url (Live365Stream *stream, GError **err)
+static gboolean
+stream_resolve (Live365Stream *stream, GError **err)
 {
+  gboolean already_resolved;
   STTransferSession *session;
-  gboolean status;
   char *url;
-  char *streaminfo;
-  char *station_id;
-  char *stream_url;
+  gboolean status;
+  char *station_id = NULL;
+
+  g_return_val_if_fail(stream != NULL, FALSE);
+
+  g_mutex_lock(stream->url_mutex);
+  already_resolved = stream->url != NULL;
+  g_mutex_unlock(stream->url_mutex);
+
+  if (already_resolved)
+    return TRUE;
 
   session = st_transfer_session_new();
   url = g_strconcat(LIVE365_ROOT, stream->url_postfix, NULL);
-  status = st_transfer_session_get(session, url, 0, NULL, &streaminfo, err);
+  status = st_transfer_session_get_by_line(session,
+					   url,
+					   0,
+					   NULL,
+					   NULL,
+					   stream_resolve_line_cb,
+					   &station_id,
+					   err);
   g_free(url);
   st_transfer_session_free(session);
 
-  if (! status)
-    return FALSE;
-  
-  station_id = streaminfo_get_station_id(streaminfo);
-  g_free(streaminfo);
-
-  if (! station_id)
+  if (status)
     {
-      g_set_error(err, 0, 0, _("stream is empty"));
-      return FALSE;
+      if (station_id)
+	{
+	  g_mutex_lock(stream->url_mutex);
+	  stream->url = ident_user && ident_session
+	    ? g_strdup_printf(LIVE365_ROOT "play/%s&session=%s&membername=%s", station_id, ident_session, ident_user)
+	    : g_strconcat(LIVE365_ROOT, "play/", station_id, NULL);
+	  g_mutex_unlock(stream->url_mutex);
+	}
+      else
+	{
+	  g_set_error(err, 0, 0, _("stream is empty"));
+	  status = FALSE;
+	}
     }
-
-  stream_url = ident_user && ident_session
-    ? g_strdup_printf(LIVE365_ROOT "play/%s&session=%s&membername=%s", station_id, ident_session, ident_user)
-    : g_strconcat(LIVE365_ROOT, "play/", station_id, NULL);
   g_free(station_id);
 
-  return stream_url;
+  return status;
 }
 
-static char *
-streaminfo_get_station_id (const char *streaminfo)
+static void
+stream_resolve_line_cb (const char *line, gpointer data)
 {
-  char **lines;
-  int i;
+  char **station_id = data;
 
-  char *station_id = NULL;
+  if (! *station_id)
+    st_re_parse(&re_stationid, line, station_id);
+}
 
-  g_return_val_if_fail(streaminfo != NULL, NULL);
-  
-  lines = g_strsplit(streaminfo, "\n", 0);
-
-  for (i = 0; lines[i]; i++)
-    if (st_re_parse(&re_stationid, lines[i], &station_id))
-      break;
-
-  g_strfreev(lines);
-
-  return station_id;
+static gboolean
+stream_resolve_cb (Live365Stream *stream, gpointer data, GError **err)
+{
+  return stream_resolve(stream, err);
 }
 
 static gboolean
@@ -323,14 +382,14 @@ stream_tune_in_cb (Live365Stream *stream,
 		   gpointer data,
 		   GError **err)
 {
-  char *url;
   gboolean status;
 
-  if (! (url = stream_get_url(stream, err)))
+  if (! stream_resolve(stream, err))
     return FALSE;
 
-  status = st_action_run("play-stream", url, err);
-  g_free(url);
+  g_mutex_lock(stream->url_mutex);
+  status = st_action_run("play-stream", stream->url, err);
+  g_mutex_unlock(stream->url_mutex);
 
   return status;
 }
@@ -340,15 +399,15 @@ stream_record_cb (Live365Stream *stream,
 		  gpointer data,
 		  GError **err)
 {
-  char *url;
   gboolean status;
 
-  if (! (url = stream_get_url(stream, err)))
+  if (! stream_resolve(stream, err))
     return FALSE;
 
-  status = st_action_run("record-stream", url, err);
-  g_free(url);
-  
+  g_mutex_lock(stream->url_mutex);
+  status = st_action_run("record-stream", stream->url, err);
+  g_mutex_unlock(stream->url_mutex);
+
   return status;
 }
 
@@ -357,7 +416,7 @@ stream_browse_cb (Live365Stream *stream,
 		  gpointer data,
 		  GError **err)
 {
-  return st_action_run("view-web", stream->station_url, err);
+  return st_action_run("view-web", stream->homepage, err);
 }
 
 static gboolean
@@ -575,9 +634,9 @@ static void
 refresh_streams_body_cb (const char *line, gpointer data)
 {
   RefreshStreamsInfo *info = data;
-
   char *sub1, *sub2, *sub3;
   char *converted = NULL;
+  char *expanded;
 
   if (! info->charset && ! strncasecmp(line, "</head>", 6))
     info->charset = g_strdup("ISO8859-1"); /* we got no charset info, use a fallback */
@@ -609,6 +668,9 @@ refresh_streams_body_cb (const char *line, gpointer data)
 	}
     }
 
+  expanded = st_sgml_ref_expand(line);
+  line = expanded;
+
   if (st_re_parse(&re_description, line, &sub1))
     {
       if (info->stream)
@@ -657,12 +719,12 @@ refresh_streams_body_cb (const char *line, gpointer data)
     {
       if (info->stream)
 	{
-	  info->stream->station_url = sub1;
+	  info->stream->homepage = sub1;
 	  info->stream->title = sub2;
 	}
       else
 	{
-	  st_notice(_("Live365: found misplaced station_url and title"));
+	  st_notice(_("Live365: found misplaced homepage and title"));
 	  g_free(sub1);
 	  g_free(sub2);
 	}
@@ -713,6 +775,7 @@ refresh_streams_body_cb (const char *line, gpointer data)
     }
 
   g_free(converted);
+  g_free(expanded);
 }
 
 static gboolean
@@ -791,7 +854,7 @@ init_handler (void)
 
   st_handler_set_label(handler, "Live365");
   st_handler_set_copyright(handler, COPYRIGHT);
-  st_handler_set_description(handler, "Live365.com - Radio Revolution");
+  st_handler_set_description(handler, "Live365 Internet Radio");
   st_handler_set_home(handler, LIVE365_ROOT);
 
   stock_categories = g_node_new(NULL);
@@ -818,8 +881,10 @@ init_handler (void)
   st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_NEW, stream_new_cb, NULL);
   st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_FIELD_GET, stream_field_get_cb, NULL);
   st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_FIELD_SET, stream_field_set_cb, NULL);
+  st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_STOCK_FIELD_GET, stream_stock_field_get_cb, NULL);
   st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_FREE, stream_free_cb, NULL);
 
+  st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_RESOLVE, stream_resolve_cb, NULL);
   st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_TUNE_IN, stream_tune_in_cb, NULL);
   st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_RECORD, stream_record_cb, NULL);
   st_handler_bind(handler, ST_HANDLER_EVENT_STREAM_BROWSE, stream_browse_cb, NULL);
@@ -827,31 +892,37 @@ init_handler (void)
   st_handler_add_field(handler, st_handler_field_new(FIELD_TITLE,
 						     _("Title"),
 						     G_TYPE_STRING,
-						     TRUE));
+						     ST_HANDLER_FIELD_VISIBLE));
   st_handler_add_field(handler, st_handler_field_new(FIELD_GENRE,
 						     _("Genre"),
 						     G_TYPE_STRING,
-						     TRUE));
+						     ST_HANDLER_FIELD_VISIBLE));
   st_handler_add_field(handler, st_handler_field_new(FIELD_DESCRIPTION,
 						     _("Description"),
 						     G_TYPE_STRING,
-						     TRUE));
+						     ST_HANDLER_FIELD_VISIBLE));
   st_handler_add_field(handler, st_handler_field_new(FIELD_BROADCASTER,
 						     _("Broadcaster"),
 						     G_TYPE_STRING,
-						     TRUE));
+						     ST_HANDLER_FIELD_VISIBLE));
   st_handler_add_field(handler, st_handler_field_new(FIELD_AUDIO,
 						     _("Audio"),
 						     G_TYPE_STRING,
-						     TRUE));
+						     ST_HANDLER_FIELD_VISIBLE));
   st_handler_add_field(handler, st_handler_field_new(FIELD_URL_POSTFIX,
 						     _("URL postfix"),
 						     G_TYPE_STRING,
-						     FALSE));
-  st_handler_add_field(handler, st_handler_field_new(FIELD_STATION_URL,
-						     _("Station URL"),
+						     0));
+  st_handler_add_field(handler, st_handler_field_new(FIELD_HOMEPAGE,
+						     _("Homepage"),
 						     G_TYPE_STRING,
-						     FALSE));
+						     ST_HANDLER_FIELD_VISIBLE
+						     | ST_HANDLER_FIELD_START_HIDDEN));
+  st_handler_add_field(handler, st_handler_field_new(FIELD_URL,
+						     _("URL"),
+						     G_TYPE_STRING,
+						     ST_HANDLER_FIELD_VISIBLE
+						     | ST_HANDLER_FIELD_START_HIDDEN));
 
   st_handlers_add(handler);
 }
@@ -864,7 +935,7 @@ plugin_init (GError **err)
   bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
   bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
 
-  if (! st_check_api_version(5, 4))
+  if (! st_check_api_version(5, 5))
     {
       g_set_error(err, 0, 0, _("API version mismatch"));
       return FALSE;
@@ -890,7 +961,7 @@ plugin_init (GError **err)
   init_handler();
 
   st_action_register("record-stream", _("Record a stream"), "xterm -hold -e streamripper %q");
-  st_action_register("view-web", _("Browse a website"), "epiphany %q");
+  st_action_register("view-web", _("Open a web page"), "epiphany %q");
   st_action_register("play-stream", _("Listen to a stream"), "xmms %q");
 
   return TRUE;
